"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.visitFnResourceBody = void 0;
const typescript_1 = __importDefault(require("typescript"));
const base_1 = require("@plutolang/base");
const imports_1 = require("./imports");
const dep_resolve_1 = require("./dep-resolve");
const constants_1 = require("./constants");
const utils_1 = require("./utils");
/**
 * Construct the FnResource and detect the access relationship within the body of FnResource.
 * @param fnNode The function expression
 * @param fnResName The function name, same with the resource name.
 */
function visitFnResourceBody(fnNode, checker, fnResName) {
    // Get the import dependencies for this function.
    const importStore = (0, imports_1.buildImportStore)(fnNode.getSourceFile());
    const importElements = (0, dep_resolve_1.resolveImportDeps)(fnNode.getSourceFile(), importStore, fnNode);
    const fnLoc = (0, utils_1.getLocationOfNode)(fnNode, 0);
    const constDepLocs = removeDuplicateLocs(detectFnAccessConst(fnNode, checker, 0));
    if (new Set(constDepLocs.concat(fnLoc).map((loc) => loc.file)).size != 1) {
        throw new Error(`Currently, Pluto only supports a single file.`);
    }
    const resourceVarInfo = {
        varName: fnResName,
        resourceConstructInfo: {
            constructExpression: constants_1.FN_RESOURCE_TYPE_NAME,
            importElements: importElements,
            locations: [fnLoc].concat(constDepLocs),
        },
    };
    // Detect the access relationship within the body of FnResource.
    // This information will be utilized to generate the permission configuration for this FnResource.
    const relatInfos = detectFnAccessResource(fnNode, checker, fnResName);
    return {
        resourceVarInfos: [resourceVarInfo],
        resourceRelatInfos: relatInfos,
    };
}
exports.visitFnResourceBody = visitFnResourceBody;
function detectFnAccessResource(fnNode, checker, fnResName) {
    const resRelatInfos = [];
    const exists = new Set();
    const checkPermission = (node) => {
        let propAccessExp;
        // Write operation, e.g. state.set(), queue.push()
        if (typescript_1.default.isExpressionStatement(node) &&
            typescript_1.default.isAwaitExpression(node.expression) &&
            typescript_1.default.isCallExpression(node.expression.expression) &&
            typescript_1.default.isPropertyAccessExpression(node.expression.expression.expression)) {
            propAccessExp = node.expression.expression.expression;
        }
        else if (typescript_1.default.isVariableStatement(node)) {
            // Read operation, e.g. state.get()
            const initExp = node.declarationList.declarations[0].initializer;
            if (initExp &&
                typescript_1.default.isAwaitExpression(initExp) &&
                typescript_1.default.isCallExpression(initExp.expression) &&
                typescript_1.default.isPropertyAccessExpression(initExp.expression.expression)) {
                propAccessExp = initExp.expression.expression;
            }
        }
        if (!propAccessExp || !(0, utils_1.isResourceVar)(propAccessExp.expression, checker)) {
            return;
        }
        const accessorName = propAccessExp.expression.getText();
        const symbol = checker.getSymbolAtLocation(propAccessExp);
        if (symbol == undefined) {
            throw new Error("The symbol of this function call is undefined: " + propAccessExp.getText());
        }
        const fnName = checker.symbolToString(symbol);
        const key = `${accessorName}-${fnName}`;
        if (exists.has(key)) {
            return;
        }
        const resRelatInfo = {
            fromVarName: fnResName,
            toVarNames: [accessorName],
            type: base_1.arch.RelatType.ACCESS,
            operation: fnName,
            parameters: [],
        };
        resRelatInfos.push(resRelatInfo);
        exists.add(key);
    };
    const fnBody = fnNode.getChildAt(fnNode.getChildCount() - 1);
    const que = [fnBody];
    while (que.length > 0) {
        const cur = que.shift();
        cur.forEachChild((child) => {
            checkPermission(child);
            que.push(child);
        });
    }
    return resRelatInfos;
}
function detectFnAccessConst(curNode, checker, depth) {
    const curFile = curNode.getSourceFile().fileName;
    const curScope = [curNode.getStart(), curNode.getEnd()];
    const locs = [];
    const fetchConstVar = (node) => {
        if (typescript_1.default.isIdentifier(node)) {
            const symbol = checker.getSymbolAtLocation(node);
            // Check if this identifier is a constant variable.
            if (!symbol || !(0, utils_1.isConstVar)(symbol) || (0, utils_1.isResourceVar)(node, checker)) {
                return;
            }
            const type = checker.getTypeOfSymbol(symbol);
            if (!type.isLiteral()) {
                throw new Error("Currently, Pluto only supports accessing constant variables with literal values that are outside the scope of a function.");
            }
            // If this is a constant variable that is defined inside the scope of this function, we can ignore it.
            const declStat = getSymbolDeclStatement(symbol);
            if ((declStat.getSourceFile().fileName.indexOf("node_modules") === -1 &&
                declStat.getSourceFile().fileName != curFile) ||
                declStat.getStart() > curScope[1] ||
                declStat.getEnd() < curScope[0]) {
                const declLoc = (0, utils_1.getLocationOfNode)(declStat, depth + 1);
                locs.push(declLoc);
                locs.push(...detectFnAccessConst(declStat, checker, depth + 1));
            }
        }
    };
    const que = [curNode];
    while (que.length > 0) {
        const cur = que.shift();
        cur.forEachChild((child) => {
            fetchConstVar(child);
            que.push(child);
        });
    }
    return locs;
}
function getSymbolDeclStatement(symbol) {
    const symbolDeclaration = symbol.valueDeclaration || symbol.declarations?.[0];
    if (symbolDeclaration == undefined) {
        throw new Error("Cannot found the declaration of symbol: " + symbol.name);
    }
    let parNode = symbolDeclaration;
    while (!typescript_1.default.isStatement(parNode)) {
        parNode = parNode.parent;
    }
    return parNode;
}
function removeDuplicateLocs(oldLocs) {
    const newLocs = [];
    for (const oldLoc of oldLocs) {
        let existed = false;
        for (const newLoc of newLocs) {
            if (oldLoc.file != newLoc.file || oldLoc.start != newLoc.start || oldLoc.end != newLoc.end) {
                continue;
            }
            existed = true;
            newLoc.depth = Math.min(newLoc.depth, oldLoc.depth);
        }
        if (!existed)
            newLocs.push(oldLoc);
    }
    return newLocs;
}
//# sourceMappingURL=visit-fn-res.js.map